import { useState, useEffect } from 'react'
import { useNavigate, useSearchParams } from 'react-router-dom'
import { Space, Typography, Form, Button, Input, DatePicker } from 'antd'
import useHotelStore from '@/stores/Hotel'
import dayjs from 'dayjs'
import { HotelList } from './HotelComponents'
import { isEmpty } from '@/utils/commons'
const { Title } = Typography

const List = () => {
  const [loading, setLoading] = useState(false)
  const [searchForm] = Form.useForm()
  const [searchByCriteria, hotelList, selectHotel] = useHotelStore((state) => [
    state.searchByCriteria,
    state.hotelList,
    state.selectHotel,
  ])

  const navigate = useNavigate()
  const [searchParams, setSearchParams] = useSearchParams()

  const hotelName = searchParams.get('hotel')
  const checkinDateString = searchParams.get('checkin')
  const checkoutDateString = searchParams.get('checkout')

  useEffect(() => {
    // 测试酒店：Hotel Skypark Kingstown Dongdaemun
    // http://localhost:5175/hotel/list?hotel=s&checkin=2024-8-20&checkout=2024-8-21
    if (
      isEmpty(hotelName) ||
      isEmpty(checkinDateString) ||
      isEmpty(checkoutDateString)
    ) {
      console.error('criteria is null')
    } else {
      searchForm.setFieldsValue({
        dataRange: [dayjs(checkinDateString), dayjs(checkoutDateString)],
        hotelName: hotelName,
      })
      setLoading(true)
      searchByCriteria(searchForm.getFieldValue()).finally(() =>
        setLoading(false),
      )
    }
  }, [])

  const handelFormFinish = () => {
    const formValue = searchForm.getFieldValue()
    setSearchParams({
      hotel: formValue.hotelName,
      checkin: formValue.dataRange[0].format('YYYY-MM-DD'),
      checkout: formValue.dataRange[1].format('YYYY-MM-DD'),
    })

    setLoading(true)
    searchByCriteria(formValue).finally(() => setLoading(false))
  }

  const handleHotelChange = (hotel) => {
    selectHotel(hotel)
    navigate(
      `/hotel/${hotel.hotel_id}/${searchForm
        .getFieldValue()
        .dataRange[0].format('YYYY-MM-DD')}/${searchForm
        .getFieldValue()
        .dataRange[1].format('YYYY-MM-DD')}`,
    )
  }

  const SearchForm = () => {
    return (
      <Form
        name="searchForm"
        form={searchForm}
        layout="inline"
        onFinish={handelFormFinish}
        onFinishFailed={() => console.info('onFinishFailed')}
        autoComplete="off"
      >
        <Form.Item
          label={'酒店名称'}
          name="hotelName"
          rules={[
            {
              required: true,
              message: '必填',
            },
          ]}
        >
          <Input placeholder={'不能为空'} />
        </Form.Item>
        <Form.Item
          label={'入住时间'}
          name="dataRange"
          allowclear="false"
          rules={[
            {
              required: true,
              message: '必填',
            },
          ]}
        >
          <DatePicker.RangePicker placeholder={['入住日期', '退房日期']} />
        </Form.Item>
        <Form.Item>
          <Button type="primary" htmlType="submit">
            搜索
          </Button>
        </Form.Item>
      </Form>
    )
  }

  return (
    <>
      <Space direction="vertical" className="w-full">
        <Title level={3}>酒店列表</Title>
        <SearchForm></SearchForm>
        <HotelList
          loading={loading}
          dataSource={hotelList}
          onChange={(h) => handleHotelChange(h)}
        ></HotelList>
      </Space>
    </>
  )
}

export default List
